<?php

require_once 'Horde/Share.php';

/**
 * Extension of the Horde_Share class for managing Ansel Galleries as
 * Horde shared objects.
 *
 * $Horde: ansel/lib/Gallery.php,v 1.89 2004/12/12 05:08:29 chuck Exp $
 *
 * @author  Chuck Hagenbuch <chuck@horde.org>
 * @author  Ben Chavet <ben@horde.org>
 * @since   Ansel 0.1
 * @package Ansel
 */
class Ansel_Gallery extends Horde_Share {

    var $_shareObject = 'DataTreeObject_Gallery';

    /**
     * Return a new gallery object
     *
     * @param string $name  The gallery's name.
     *
     * @return object DataTreeObject_Gallery a new gallery object.
     */
    function &newGallery($name)
    {
        if (empty($name)) {
            return PEAR::raiseError(_("Gallery names must be non-empty"));
        }
        $gallery = &new DataTreeObject_Gallery($name);
        $gallery->setShareOb($this);
        return $gallery;
    }

    /**
     * Return a new image object
     *
     * @param string $name           The image's name
     *
     * @param optional array $image  Array containing parameters to send to
     *                               the new image constructor
     *
     * @return object DataTreeObject_Image a new image object.
     */
    function &newImage($name, $image = null)
    {
        if (empty($name)) {
            return PEAR::raiseError(_("Image names must be non-empty"));
        }
        $image = &new DataTreeObject_Image($name, $image);
        $image->setImageOb($this);
        return $image;
    }

    /**
     * Return a DataTreeObject_Gallery object corresponding to the
     * given name/ID.
     *
     * @param string $name  The name or ID of the gallery to retrieve.
     *
     * @return  The gallery object corresponding to the given name.
     */
    function &getGallery($id)
    {
        /* cache of previous retrieved galleries */
        static $galleryCache;

        if (!is_array($galleryCache)) {
            $galleryCache = array();
        }

        if (!isset($galleryCache[$id])) {
            $galleryCache[$id] = $this->_datatree->getObjectById($id, 'DataTreeObject_Gallery');
            if (!is_a($galleryCache[$id], 'PEAR_Error')) {
                $galleryCache[$id]->setShareOb($this);
            }
        }

        return $galleryCache[$id];
    }

    /**
     * Returns a DataTreeObject_Image object corresponding to the
     * given name.
     *
     * @param string $name  The name of the image to retrieve.
     *
     * @return  The image object corresponding to the given name.
     */
    function &getImage($id)
    {
        /* Cache of previous retrieved images. */
        static $imageCache;

        if (!is_array($imageCache)) {
            $imageCache = array();
        }

        if (!isset($imageCache[$id])) {
            $imageCache[$id] = $this->_datatree->getObjectById($id, 'DataTreeObject_Image');
            if (!is_a($imageCache[$id], 'PEAR_Error')) {
                $imageCache[$id]->setImageOb($this);
            }
        }
        return $imageCache[$id];
    }

    function getName($id)
    {
        return $this->_datatree->getName($id);
    }

    /**
     * Returns a list of images maching the specified criteria.
     */
    function searchAllImages($criteria)
    {
        return $this->_datatree->getByAttributes($criteria, DATATREE_ROOT, true, false);
    }

    /**
     * Adds a new DataTreeObject_Gallery to the object backend
     *
     * @param object DataTreeObject_Gallery $gallery  The gallery to add
     */
    function addGallery($gallery)
    {
        if (!is_a($gallery, 'DataTreeObject_Gallery')) {
            return PEAR::raiseError('Galleries must be DataTreeObject_Gallery objects or extend that class.');
        }
        return $this->_datatree->add($gallery);
    }

    /**
     * Adds a new DataTreeObject_Image to the datatree backend
     *
     * @param object DataTreeObject_Image $image  The image to add
     */
    function addImage($image)
    {
        if (!is_a($image, 'DataTreeObject_Image')) {
            return PEAR::raiseError('Images must be DataTreeObject_Image objects or extend that class.');
        }
        return $this->_datatree->add($image);
    }

    /**
     * Removes a DataTreeObject_Gallery from the datatree backend.
     *
     * @param object DataTreeObject_Gallery$gallery  The gallery to remove.
     *
     * @param optional boolean $force [default = false] Force to remove
     *                         every child.
     */
    function removeGallery($gallery, $force = false)
    {
        if (!is_a($gallery, 'DataTreeObject_Gallery')) {
            return PEAR::raiseError('Galleries must be DataTreeObject_Gallery objects or extend that class.');
        }
        return $this->_datatree->remove($gallery, $force);
    }

    /**
     * Remove an image from the datatree backend.
     *
     * @param object DataTreeObject_Image $image  The image to remove.
     */
    function removeImage($image)
    {
        if (!is_a($image, 'DataTreeObject_Image')) {
            return PEAR::raiseError('Images must be DataTreeObject_Image objects or extend that class.');
        }
        return $this->_datatree->remove($image, false);
    }

    /**
     * Check if a gallery or image exists in the system.
     *
     * @param string $datatreeObject  The DataTreeObject to check.
     *
     * @return boolean true if the group exists, false otherwise.
     */
    function exists($datatreeObject)
    {
        return $this->_datatree->exists($datatreeObject);
    }

}

/**
 * Extension of the DataTreeObject_Share class for storing Galleries
 * as Horde shared objects.
 *
 * @author  Mike Cochrane <mike@graftonhall.co.nz>
 * @author  Ben Chavet <ben@horde.org>
 * @since   Ansel 0.1
 * @package Ansel
 */
class DataTreeObject_Gallery extends DataTreeObject_Share {

    /**
     * The DataTreeObject_Gallery constructor. Calls the parent constructor
     * and sets the default image dimensions.
     *
     * @param string $name  The name of the gallery
     */
    function DataTreeObject_Gallery($name)
    {
        parent::DataTreeObject_Share($name);

        // Make sure default height and width settings are present.
        global $conf;
        $this->data['thumbnail_width'] = $conf['thumbnail']['width'];
        $this->data['thumbnail_height'] = $conf['thumbnail']['height'];
        $this->data['image_width'] = $conf['screen']['width'];
        $this->data['image_height'] = $conf['screen']['height'];

        // Initialize gallery info variables.
        $this->data['default'] = DATATREE_ROOT;
        $this->data['default-type'] = 'auto';
        $this->data['num-images'] = 0;
        $this->data['date-created'] = time();
        $this->data['last-modified'] = time();
        $this->data['category'] = '';
    }

    /**
     * Add an image to this gallery
     *
     * @param array $image  The image to add.  Expected keys include
     *                      'description', and 'data'. Optional keys include
     *                      'filename' and 'type'
     *
     * @return  The datatree name for the new image.
     */
    function addImage($image)
    {
        global $conf;

        // Make sure we have a filename for the image.
        if (!isset($image['filename'])) {
            $image['filename'] = 'Untitled';
        }

        // Create image object and add it to the backend.
        $imageOb = &$this->_shareOb->newImage(md5(microtime()), $image);
        $imageOb->set('gallery', $this->getId());
        $this->_shareOb->addImage($imageOb);
        $this->setImageOrder($imageOb->getId(), $this->get('num_images'));
        if (!empty($image['data'])) {
            $result = $imageOb->writeData();
            if (is_a($result, 'PEAR_Error')) {
                return $result;
            }
        }

        require_once ANSEL_BASE . '/lib/Exif.php';
        Ansel_ImageData::setExifData($imageOb);
        $orientation = $imageOb->get('Orientation');
        if ($orientation !== null && $orientation != 1) {
            switch ($orientation) {
            case 2:
                $imageOb->mirror();
                break;

            case 3:
                $imageOb->rotate(180);
                break;

            case 4:
                $imageOb->mirror();
                $imageOb->rotate(180);
                break;

            case 5:
                $imageOb->flip();
                $imageOb->rotate(90);
                break;

            case 6:
                $imageOb->rotate(90);
                break;

            case 7:
                $imageOb->mirror();
                $imageOb->rotate(90);
                break;

            case 8:
                $imageOb->rotate(270);
                break;
            }

            $imageOb->set('Orientation', 1);
            $imageOb->update();
            $imageOb->updateData($imageOb->raw());
        }

        // Increment image count.
        $this->data['num-images']++;

        // Change the default image if it has been configured
        // automatically.
        if ($this->data['default-type'] == 'auto') {
            $this->data['default'] = $imageOb->getId();
        }

        // Update the modified flag.
        $this->data['last-modified'] = time();

        // Store changes to the gallery.
        $this->update();

        // Return the ID of the new image.
        return $imageOb->getId();
    }

    function removeImage($image, $delete = false)
    {
        // Make sure $image is a DataTreeObject_Image; if not,
        // try loading it from the datatree
        if (!is_a($image, 'DataTreeObject_Image')) {
            $img = $this->_shareOb->getImage($image);
            if (is_a($img, 'PEAR_Error')) {
                return $img;
            }
            $image = $img;
        }

        // Make sure the image is in this gallery.
        if ($this->getId() != $image->get('gallery')) {
            return false;
        }

        // Change gallery info.
        $this->data['num-images']--;
        if ($this->data['default'] == $image->getId()) {
            $this->data['default'] = DATATREE_ROOT;
            $this->data['default-type'] = 'auto';
        }

        // Update the modified flag
        $this->data['last-modified'] = time();

        $this->update();

        // Remove image.
        if ($delete) {
            $image->delete();
        } else {
            $image->set('gallery', DATATREE_ROOT);
        }

        return true;
    }

    /**
     * Set one of the attributes of the object.
     *
     * @param string  $attribute  The attribute to set.
     * @param mixed   $value      The value for $attribute.
     * @param boolean $update     Determines whether the backend should
     *                            be updated or not
     */
    function set($attribute, $value, $update = false)
    {
        parent::set($attribute, $value);
        if ($update) {
            $this->update();
        }
    }

    /**
     * Updates the backend to with the current values of this gallery
     */
    function update()
    {
        return $this->_shareOb->_datatree->updateData($this);
    }

    /**
     * Renames this gallery.
     *
     * @param string $newName  The new name for this gallery.
     */
    function rename($newName)
    {
        return $this->_shareOb->_datatree->rename($this, $newName);
    }

    /**
     * Explicitly set the order for a photo in this gallery.
     *
     * @param integer $id     The id of the image.
     * @param integer $order  The new order position for the image.
     */
    function setImageOrder($id, $order)
    {
        return $this->_shareOb->_datatree->setOrder($id, $order);
    }

    /**
     * Returns the parents of this gallery in flat array.
     */
    function getParents()
    {
        $parents = array();
        $gallery = $this;
        while (is_a($gallery = $gallery->getParent(), 'DataTreeObject_Gallery')) {
            $parents[] = $gallery->getId();
        }
        return array_reverse($parents);
    }

    /**
     * Returns the owner of this gallery.
     *
     * @return Identity object for the owner of this gallery.
     */
    function getOwner()
    {
        return $this->_shareOb->getIdentityByShare($this);
    }

    /**
     * Lists the images in this gallery.
     *
     * @param integer $from     The image to start fetching
     * @param integer $count     The numer of images to return
     */
    function listImages($from = 0, $count = 0)
    {
        $criteria = array('AND' => array(
                        array('field' => 'name',
                              'op' => '=',
                              'test' => 'gallery'),
                        array('field' => 'value',
                              'op' => '=',
                              'test' => $this->getId())));
        return $this->_shareOb->_datatree->getByAttributes($criteria, DATATREE_ROOT, true, true, $from, $count);
    }

    /**
     * Returns the number of images in this gallery and all sub-galleries.
     */
    function numImages()
    {
        $galleries = $this->_shareOb->listShares(Auth::getAuth(), PERMS_SHOW, '', $this->getName());
        $count = $this->get('num-images');
        foreach ($galleries as $galleryId => $gallery) {
            $count += $gallery->get('num-images');
        }
        return $count;
    }

    /**
     * Returns the default image for this gallery.
     */
    function getDefaultImage()
    {
        if (isset($this->data['default'])) {
            if ($this->data['default'] == DATATREE_ROOT) {
                if ($this->data['num-images'] > 0) {
                    $this->_findDefaultImage();
                    return $this->data['default'];
                } elseif ($this->_shareOb->hasChildren($this)) {
                    $galleries = $this->_shareOb->listShares(Auth::getAuth(), PERMS_SHOW, '', $this->getName());
                    foreach ($galleries as $galleryId => $gallery) {
                        if ($gallery->getDefaultImage()) {
                            return $gallery->getDefaultImage();
                        }
                    }
                }
            } else {
                return $this->data['default'];
            }
        }

        return false;
    }

    function _findDefaultImage()
    {
        $keys = array_keys($this->listImages());
        $this->set('default', $keys[count($keys) - 1], true);
        $this->set('default-type', 'auto', true);
    }

    /**
     * Map this object's attributes from the data array into a format that
     * we can store in the attributes storage backend.
     *
     * @return array  The attributes array
     */
    function _toAttributes()
    {
        // Default to no attributes.
        $attributes = parent::_toAttributes(true);

        foreach ($this->data as $key => $value) {
            if ($key == 'perm') {
                continue;
            } else {
                $attributes[] = array('name' => $key,
                                      'key' => '',
                                      'value' => $value);
            }
        }

        return $attributes;
    }

    /**
     * Take in a list of attributes from the backend and map it to our
     * internal data array.
     *
     * @param array $attributes  The list of attributes from the
     *                           backend (attribute name, key, and value).
     */
    function _fromAttributes($attributes)
    {
        parent::_fromAttributes($attributes, true);

        foreach ($attributes as $attr) {
            if (substr($attr['name'], 0, 4) == 'perm') {
                continue;
            } else {
                $this->data[$attr['name']] = $attr['value'];
            }
        }
    }

}

/**
 * Extension of the DataTreeObject class for storing Images in as Horde
 * datatree objects.
 *
 * @author  Ben Chavet <ben@horde.org>
 * @package Ansel
 */
class DataTreeObject_Image extends DataTreeObject {

    var $_vfs;
    var $_imageOb;
    var $_image;
    var $_dirty;
    var $_loaded = array();
    var $_data = array();

    /**
     * DataTreeObject_Image constructor.  Creates VFS reference and
     * fills in any supplied attributes.
     *
     * @param string $name  Name of this image
     *
     * @param optional array $image Optional array of image attributes.
     *                              If supplied, required keys are 'filename',
     *                              'description', and 'data'.  Optional keys
     *                              are 'date' and 'type'
     */
    function DataTreeObject_Image($name, $image = null)
    {
        parent::DataTreeObject($name);

        global $conf;
        $this->_vfs = &VFS::singleton($conf['vfs']['type'], Horde::getDriverConfig('vfs', $conf['vfs']['type']));
        if ($logger = &Horde::getLogger()) {
            $this->_vfs->setLogger($logger, $GLOBALS['conf']['log']['priority']);
        }

        $this->_image = &Ansel::getImageObject();
        $this->_image->reset();

        $this->data['gallery'] = DATATREE_ROOT;

        if (is_array($image)) {
            $this->data['filename'] = $image['filename'];
            $this->data['description'] = $image['description'];
            $this->data['date'] = empty($image['date']) ? time() : $image['date'];

            if (!empty($image['type'])) {
                $this->data['type'] = $image['type'];
            }

            if (!empty($image['data'])) {
                $this->_data['full'] = $image['data'];
            }
        } else {
            $this->data['filename'] = 'Untitled';
            $this->data['description'] = '';
            $this->data['date'] = time();
        }
    }

    /**
     * Associates an Ansel_Gallery object with this image.
     *
     * @param object Ansel_Gallery $imageOb  The Ansel_Gallery object.
     */
    function setImageOb(&$imageOb)
    {
        $this->_imageOb = $imageOb;
    }


    /**
     * Return the vfs path for this gallery.
     *
     * @return string  The vfs path for this gallery.
     */
    function vfspath($view = 'full')
    {
        return '.horde/ansel/' . substr($this->getId(), -2) . '/' . $view;
    }

    /**
     * Set one of the attributes of the object.
     *
     * @param string   The attribute to set.
     * @param mixed    The value for $attribute.
     * @param boolean  Determines whether the backend should
     *                 be updated or not
     */
    function set($attribute, $value, $update = false)
    {
        parent::set($attribute, $value);
        if ($update) {
            $this->update();
        }
    }

    function getId()
    {
        return $this->_imageOb->_datatree->getId($this);
    }

    /**
     * Returns the file name of this image as used in the VFS backend.
     *
     * @return string  This image's VFS file name.
     */
    function getVFSName($view)
    {
        $vfsname = $this->getId();

        $type = $view == 'full' ? $this->get('type') : 'image/' . $GLOBALS['conf']['image']['type'];
        require_once 'Horde/MIME/Magic.php';
        if ($ext = MIME_Magic::MIMEToExt($type)) {
            $vfsname .= '.' . $ext;
        }

        return $vfsname;
    }

    /**
     * Updates the datatree backend with the current values of this
     * image.
     */
    function update()
    {
        return $this->_imageOb->_datatree->updateData($this);
    }

    /**
     * Renames this image.
     *
     * @param string $newName  The new name for this image.
     */
    function rename($newName)
    {
        return $this->_imageOb->_datatree->rename($this, $newName);
    }

    /**
     * Loads the given view into memory.
     *
     * @param optional string $view  Which view to load.
     */
    function load($view = 'full')
    {
        // If we've already loaded the data, just return now.
        if (!empty($this->_loaded[$view])) {
            return;
        }

        $result = $this->createView($view);
        if (is_a($result, 'PEAR_Error')) {
            return $result;
        }

        // If createView() had to resize the full image, we've already
        // loaded the data, so return now.
        if (!empty($this->_loaded[$view])) {
            return;
        }

        // We've definitely successfully loaded the image now.
        $this->_loaded[$view] = true;

        // Get the VFS info.
        $vfspath = $this->vfspath($view);

        // Read in the requested view.
        $data = $this->_vfs->read($vfspath, $this->getVFSName($view));
        if (is_a($data, 'PEAR_Error')) {
            return $data;
        }

        $this->_data[$view] = $data;
        $this->_image->loadString($vfspath . '/' . $this->getId(), $data);
    }

    /**
     * Creates and caches the given view.
     *
     * @param string view  Which view to create.
     */
    function createView($view)
    {
        // Get the VFS info.
        $vfspath = $this->vfspath($view);

        $check = $this->_vfs->exists($vfspath, $this->getVFSName($view));
        if ($check) {
            return $check;
        }

        $data = $this->_vfs->read($this->vfspath('full'), $this->getVFSName('full'));
        if (is_a($data, 'PEAR_Error')) {
            return $data;
        }
        if (is_a($result = $this->_image->loadString($this->vfspath('full') . '/' . $this->getId(), $data), 'PEAR_Error')) {
            return $result;
        }

        switch ($view) {
        case 'screen':
            // Create and cache the screen size image.
            $gallery = $this->_imageOb->getGallery($this->data['gallery']);
            $dimensions = $this->_image->getDimensions();
            $this->_image->resize(min($gallery->get('image_width'), $dimensions['width']),
                                  min($gallery->get('image_height'), $dimensions['height']),
                                  true);

            // Store the data, since we have it loaded now.
            $this->_data[$view] = $this->_image->raw();
            $this->_image->loadString($vfspath . '/' . $this->getId(), $this->_data[$view]);
            $this->_loaded[$view] = true;

            // Write the resized image to the backend.
            $this->_vfs->writeData($vfspath, $this->getVFSName($view), $this->_data[$view], true);
            break;

        case 'thumb':
            // Create and cache the thumbnail.
            $gallery = $this->_imageOb->getGallery($this->data['gallery']);
            $dimensions = $this->_image->getDimensions();
            $this->_image->resize(min($gallery->get('thumbnail_width'), $dimensions['width']),
                                  min($gallery->get('thumbnail_height'), $dimensions['height']),
                                  true);

            // Store the data, since we have it loaded now.
            $this->_data[$view] = $this->_image->raw();
            $this->_image->loadString($vfspath . '/' . $this->getId(), $this->_data[$view]);
            $this->_loaded[$view] = true;

            // Write the resized image to the backend.
            $this->_vfs->writeData($vfspath, $this->getVFSName($view), $this->_data[$view], true);
            break;

        case 'mini':
            // Create and cache the mini.
            $gallery = $this->_imageOb->getGallery($this->data['gallery']);
            $dimensions = $this->_image->getDimensions();
            $this->_image->resize(min(50, $dimensions['width']), min(50, $dimensions['height']), true);

            // Store the data, since we have it loaded now.
            $this->_data[$view] = $this->_image->raw();
            $this->_image->loadString($vfspath . '/' . $this->getId(), $this->_data[$view]);
            $this->_loaded[$view] = true;

            // Write the resized image to the backend.
            $this->_vfs->writeData($vfspath, $this->getVFSName($view), $this->_data[$view], true);
            break;
        }
    }

    /**
     * Writes the current data to vfs.  Any cached images are deleted
     * so they can be regenerated from the new data when necessary.
     */
    function writeData()
    {
        // Delete cache and write data.
        $this->deleteCache();
        $this->_dirty = false;

        return $this->_vfs->writeData($this->vfspath('full'), $this->getVFSName('full'), $this->_data['full'], true);
    }

    /**
     * Removes this image from the VFS and DataTree backends.
     */
    function delete()
    {
        // Delete cached files from VFS.
        $this->deleteCache();

        // Delete original image from VFS.
        $this->_vfs->deleteFile($this->vfspath('full'), $this->getVFSName('full'));

        // Remove from DataTree backend.
        $this->_imageOb->removeImage($this);
    }

    /**
     * Deletes the specified cache file. If none is specified, deletes
     * all of the cache files.
     *
     * @param optional string $view  Which cache file to delete.
     */
    function deleteCache($view = 'all')
    {
        // Delete cached screen image.
        if ($view == 'all' || $view == 'screen') {
            $this->_vfs->deleteFile($this->vfspath('screen'), $this->getVFSName('screen'));
        }

        // Delete cached thumbnail.
        if ($view == 'all' || $view == 'thumb') {
            $this->_vfs->deleteFile($this->vfspath('thumb'), $this->getVFSName('thumb'));
        }

        // Delete cached mini image.
        if ($view == 'all' || $view == 'mini') {
            $this->_vfs->deleteFile($this->vfspath('mini'), $this->getVFSName('mini'));
        }
    }

    /**
     * Returns the raw data for the given view.
     *
     * @param optional string $view  Which view to return.
     */
    function raw($view = 'full')
    {
        $this->load($view);
        if ($this->_dirty) {
            return $this->_image->raw();
        } else {
            return $this->_data[$view];
        }
    }

    /**
     * Change the image data. Deletes old cache and writes the new
     * data to the VFS.
     *
     * @param string $data  The new data for this image.
     */
    function updateData($data)
    {
        if (is_a($data, 'PEAR_Error')) {
            return $data;
        }

        // Delete old cached data and write the full data.
        $this->deleteCache();
        return $this->_vfs->writeData($this->vfspath('full'), $this->getVFSName('full'), $data, true);
    }

    /**
     * Sends the correct HTTP headers to the browser to download this image.
     *
     * @param string $view  The view to download.
     */
    function downloadHeaders($view = 'full')
    {
        global $browser, $conf;

        $filename = $this->get('filename');
        if ($view != 'full') {
            require_once 'Horde/MIME/Magic.php';
            if ($ext = MIME_Magic::MIMEToExt('image/' . $conf['image']['type'])) {
                $filename .= '.' . $ext;
            }
        }

        $browser->downloadHeaders($filename);
    }

    /**
     * Display the requested view.
     *
     * @param string $view  Which view to display.
     */
    function display($view = 'full')
    {
        if ($view == 'full' && !$this->_dirty) {
            if (is_a($data = $this->_vfs->read($this->vfspath('full'), $this->getVFSName('full')), 'PEAR_Error')) {
                return $data;
            }
            echo $data;
            return;
        }
        if (is_a($result = $this->load($view), 'PEAR_Error')) {
            return $result;
        }
        return $this->_image->display();
    }

    /**
     * Wraps the given fiew into a file.
     *
     * @param string $view  Which view to wrap up.
     */
    function toFile($view = 'full')
    {
        $this->load($view);
        return $this->_image->toFile($this->_dirty ? false : $this->_data[$view]);
    }

    /**
     * Returns the dimensions of the given view.
     *
     * @param string $view  Which file to check dimensions for.
     */
    function getDimensions($view = 'full')
    {
        $this->load($view);
        return $this->_image->getDimensions();
    }

    /**
     * Rotates the image.
     *
     * @param integer $angle  What angle to rotate the image by.
     */
    function rotate($angle)
    {
        $this->load();
        $this->_dirty = true;
        return $this->_image->rotate($angle);
    }

    /**
     * Crops the image.
     *
     * @param integer $x1  First X coordinate.
     * @param integer $y1  First Y coordinate.
     * @param integer $x2  Second X coordinate.
     * @param integer $y2  Second Y coordinate.
     */
    function crop($x1, $y1, $x2, $y2)
    {
        $this->load();
        $this->_dirty = true;
        return $this->_image->crop($x1, $y1, $x2, $y2);
    }

    /**
     * Converts the image to grayscale.
     */
    function grayscale()
    {
        $this->load();
        $this->_dirty = true;
        return $this->_image->grayscale();
    }

    /**
     * Watermarks the image.
     *
     * @param optional string $watermark  String to use as the watermark.
     */
    function watermark($watermark = null, $halign = null, $valign = null, $font = null)
    {
        if (empty($watermark)) {
            require_once 'Horde/Identity.php';
            $identity = &Identity::singleton();
            $name = $identity->getValue('fullname');
            if (empty($name)) {
                $name = Auth::getAuth();
            }
            $watermark = sprintf(_("(c) %s %s"), date('Y'), $name);
        }

        $this->load();
        $this->_dirty = true;
        return $this->_image->watermark($watermark, $halign, $valign, $font);
    }

    /**
     * Flips the image.
     */
    function flip()
    {
        $this->load();
        $this->_dirty = true;
        return $this->_image->flip();
    }

    /**
     * Mirrors the image.
     */
    function mirror()
    {
        $this->load();
        $this->_dirty = true;
        return $this->_image->mirror();
    }

    /**
     * Map this object's attributes from the data array into a format
     * that we can store in the attributes storage backend.
     *
     * @return array  The attributes array.
     */
    function _toAttributes()
    {
        foreach($this->data as $key => $value) {
            if ($key == 'perm') {
                continue;
            } else {
                $attributes[] = array('name' => $key,
                                      'key' => '',
                                      'value' => $value);
            }
        }

        return $attributes;
    }

    /**
     * Take in a list of attributes from the backend and map it to our
     * internal data array.
     *
     * @param array $attributes  The list of attributes from the
     *                           backend (attribute name, key, and value).
     */
    function _fromAttributes($attributes)
    {
        foreach ($attributes as $attr) {
            if (substr($attr['name'], 0, 4) == 'perm') {
                continue;
            } else {
                $this->data[$attr['name']] = $attr['value'];
            }
        }
    }

}
